from __future__ import with_statement
from framework.latentmodule import LatentModule
from framework.ui_elements.EventWatcher import EventWatcher
import random
import time

class AdvCommScheduler(LatentModule):
    """
    A scheduler (almost complete subtask) that presents a series of distractor messages (some of which have a callsign and others having no callsign),    
    with occasional targeted messages (aimed at the subject's assigned callsign), a fraction of which is accompanied by yes/no questions. The targeted 
    questions come in bursts that are called "situations" or "developments" here.
    """    
    def __init__(self,
                 # output environment
                 presenter,                                 # stimulus presenter to use
                 rewardlogic,                               # the reward handler                 
                 focused=True,                              # whether this object is in the user's focus
                 markerbase=300,                            # base marker offset
                 
                 # content control           
                 targetsign=None,                           # the user's assigned callsign (if None, will be randomly selected)
                 callsigns='callsigns.txt',                 # source file containing a list of call signs
                 numcallsigns=6,                            # subset of callsigns to use                 
                 commands='sentences_with_answers.txt',   # source file containing a list of actionable commands (sentences and assoc. questions)
                 distractors='distractor_sentences.txt',    # source file containing a list of distractor sentences

                 # probabilities & timing control
                 lull_time = lambda: random.uniform(30,90),      # duration of lulls, in seconds (drawn per lull)
                 situation_time = lambda: random.uniform(10,35), # duration of developing situations, in seconds (drawn per situation)
                 clearafter = 4,                                 # clear presenter this many seconds after message display
                 message_interval = lambda: random.uniform(4,6), # message interval, in s (drawn per message)
                 other_callsign_fraction = lambda: random.uniform(0.6,0.75),      # fraction of messages that are for other callsigns (out of all messages presented) (drawn per situation)
                 no_callsign_fraction = lambda: random.uniform(0.25,0.35),        # fraction, out of the messages for "other callsigns", of messages that have no callsign (drawn per situation)
                 time_fraction_until_questions = lambda: random.uniform(0.5,0.8), # the fraction of time into the situation until the first question comes up (drawn per situation)
                                                                                  # in the tutorial mode, this should probably be close to zero
                 questioned_fraction = lambda: random.uniform(0.6,0.8),           # fraction of targeted messages that incur questions
                                  
                 # end conditions
                 end_trials=100000,                         # number of trials to produce
                 end_timeout=100000,                        # lifetime of this stream, in seconds
                 
                 # response control
                 events=['y','n','s'],                      # event names that encode yes, no, and skip
                 response_timeout = 3,                      # response timeout...                    
                 loss_incorrect=-2,                         # amount of loss incurred when incorrectly answering
                 gain_correct=1,                            # amount of reward gained when answering correctly 
                 loss_skipped=-1,                           # amount of loss incurred when admitting a miss
                 loss_missed=-2,                            # amount of loss incurred when missing the question (and basically the sentence, too)
                 
                 # bci features
                 callback_delay=0.8,                        # query the BCI this many seconds after a "targeted/important" message was displayed 
                 callback_func=None,                        # call this callback function to do it
                 
                 # question counter
                 num_question=None                          # the current question index from where we continue (otherwise assumed 0)
                 ):
               
        LatentModule.__init__(self)
        self.presenter = presenter
        self.rewardlogic = rewardlogic
        self.focused = focused
        self.markerbase = markerbase
        self.targetsign = targetsign

        self.callsigns = []
        with open('media\\'+callsigns,'r') as f:
            for line in f:
                self.callsigns.append(line.strip())
        self.callsigns = self.callsigns[:numcallsigns]               
        # load commands
        self.sentences = [] # actionable sentences
        self.questions = [] # questions about those sentences
        self.responses = [] # correct answers to the questions
        with open('media\\'+commands,'r') as f:
            for line in f:
                parts = line.split('|')
                try:
                    self.sentences.append(parts[0].strip())
                    self.questions.append(parts[1].strip())
                    self.responses.append(parts[2].strip())
                except:
                    pass
                # permute the order of these things
        #order = range(len(self.sentences))
        #random.shuffle(order)
        #self.sentences = [self.sentences[k] for k in order]
        #self.questions = [self.questions[k] for k in order]
        #self.responses = [self.responses[k] for k in order]
        
        # load distractors
        with open('media\\'+distractors,'r') as f:
            self.distractors = f.readlines()
        random.shuffle(self.distractors)

        # timing parameters    
        self.lull_time = lull_time
        self.situation_time = situation_time
        self.message_interval = message_interval
        self.other_callsign_fraction = other_callsign_fraction
        self.no_callsign_fraction = no_callsign_fraction
        self.time_fraction_until_questions = time_fraction_until_questions
        self.questioned_fraction = questioned_fraction  

        self.end_trials = end_trials
        self.end_timeout = end_timeout

        self.response_timeout = response_timeout
        self.loss_incorrect=loss_incorrect
        self.loss_missed=loss_missed
        self.loss_skipped=loss_skipped
        self.gain_correct=gain_correct
        
        self.events=events
        self.clearafter = clearafter
        
        self.callback_delay = callback_delay
        self.callback_func = callback_func
        
        if num_question is None:
            self.num_question = [0]
        else:
            self.num_question = num_question
        
    def run(self):        
        # pick a random callsign if necessary
        if self.targetsign is None:
            signidx = random.choice(xrange(len(self.callsigns)))
            self.marker(self.markerbase+20+signidx)
            self.targetsign = self.callsigns[signidx];
            
        # remove the targetsign from the callsigns
        self.callsigns.remove(self.targetsign)

        self.sleep(self.message_interval())
        t_terminate = time.time() + self.end_timeout

        while True:
            # begin a lull sequence
            lull_duration = self.lull_time();
            no_callsign_fraction = self.no_callsign_fraction()
            t_end = time.time() + lull_duration
            while time.time() < t_end:
                # check termination condition
                if time.time() > t_terminate:
                    break

                # message for another callsign
                if random.random() < no_callsign_fraction:
                    # has no callsign
                    self.marker(self.markerbase+1)
                    self.presenter.clear()
                    self.presenter.submit_wait(random.choice(self.distractors),self,clearafter=self.clearafter)
                else:
                    # for another callsign
                    self.marker(self.markerbase+2)
                    self.presenter.clear()
                    self.presenter.submit_wait(self.substitute(random.choice(self.distractors),random.choice(self.callsigns)),self,clearafter=self.clearafter)                             
                # wait for the message interval
                self.sleep(self.message_interval())
                    
            # begin an action sequence
            situation_time = self.situation_time()
            t_end = time.time() + situation_time 
            other_callsign_fraction = self.other_callsign_fraction()
            no_callsign_fraction = self.no_callsign_fraction()
            time_fraction_until_questions = self.time_fraction_until_questions() 
            t_beginquestions = time.time() + situation_time * time_fraction_until_questions
            questioned_fraction = self.questioned_fraction()
            
            while time.time() < t_end:
                # check termination condition
                if time.time() > t_terminate:
                    break

                if random.random() < other_callsign_fraction:
                    # message for another callsign
                    if random.random() < no_callsign_fraction:
                        # has no callsign
                        self.marker(self.markerbase+3)
                        self.presenter.clear()
                        self.presenter.submit_wait(random.choice(self.distractors),self,clearafter=self.clearafter)
                        self.sleep(self.message_interval())
                    else:
                        # for another callsign
                        self.marker(self.markerbase+4)
                        self.presenter.clear()
                        self.presenter.submit_wait(self.substitute(random.choice(self.distractors),random.choice(self.callsigns)),self,clearafter=self.clearafter)
                        self.sleep(self.message_interval())
                else:
                    # message for the current callsign
                    if time.time() < t_beginquestions:
                        # no question asked
                        self.marker(self.markerbase+5)
                        self.presenter.clear()
                        self.presenter.submit_wait(self.substitute(random.choice(self.distractors),self.targetsign),self,clearafter=self.clearafter)
                        self.sleep(self.callback_delay)
                        if self.callback_func is not None:
                            self.callback_func()                        
                        self.sleep(self.message_interval()-self.callback_delay)
                    else:
                        if random.random() >= questioned_fraction: 
                            # no question asked
                            self.marker(self.markerbase+5)
                            self.presenter.clear()
                            self.presenter.submit_wait(self.substitute(random.choice(self.distractors),self.targetsign),self,clearafter=self.clearafter)
                            self.sleep(self.callback_delay)
                            if self.callback_func is not None:
                                self.callback_func()                        
                            self.sleep(self.message_interval()-self.callback_delay)
                        else:
                            # first present the sentence
                            self.marker(self.markerbase+6)
                            self.presenter.clear()
                            self.presenter.submit_wait(self.substitute(self.sentences[self.num_question[0]],self.targetsign),self,clearafter=self.clearafter)
                            # wait for the message interval
                            self.sleep(self.callback_delay)
                            if self.callback_func is not None:
                                self.callback_func()                        
                            self.sleep(self.message_interval()-self.callback_delay)
                            # present the question
                            self.marker(self.markerbase+7 if self.focused else self.markerbase+8)
                            self.presenter.clear()
                            self.presenter.submit_wait(self.substitute(self.questions[self.num_question[0]],self.targetsign),self,clearafter=self.clearafter)
                            correct_answer = self.responses[self.num_question[0]]
                            self.num_question[0] += 1                        
                            duration = self.message_interval()                        
                            if self.focused:
                                # ... wait for a response
                                begin_wait = time.time()
                                response = self.waitfor_multiple(self.events, self.response_timeout)
                                if response == None:
                                    # missed
                                    self.marker(self.markerbase + 11)
                                    self.rewardlogic.score_event(self.loss_missed,True)
                                elif response[0] == self.events[2]:
                                    # skipping the answer
                                    self.marker(self.markerbase + 9)
                                    self.rewardlogic.score_event(self.loss_skipped)
                                elif (correct_answer=='y' and response[0]==self.events[0]) or (correct_answer=='n' and response[0]==self.events[1]): 
                                    # correct answer
                                    self.marker(self.markerbase + 10)
                                    self.rewardlogic.score_event(self.gain_correct)
                                else:
                                    # incorrect answer
                                    self.marker(self.markerbase + 12)
                                    self.rewardlogic.score_event(self.loss_incorrect)
                                # wait for the rest of the duration
                                self.sleep(duration - (time.time()-begin_wait))
                            else:
                                # just wait for the msg interval
                                self.sleep(duration)


    def substitute(self,command,callsign):
        """Substitute a callsign into a command."""
        if command.find(' *')>0:
            # placeholder not at the beginning
            command.replace('*',callsign.lower())
        elif command.find('*')>0:
            # placeholder at the beginning
            command.replace('*',callsign)
        else:
            # no placeholder, prepend callsign
            command = callsign + '; ' + command
        return command
        