from __future__ import with_statement
from framework.basicstimuli import BasicStimuli
import time

#
# This is a reward logic that produces audio feedback. 
# Also supports score logging.
#

class AudioRewardLogic(BasicStimuli):
    """
    This class does all the reward things (counts the score and plays sounds when the score is to be updated).
    See bottom of this file for the marker table.
    """
    def __init__(self,
                 initial_score=10,                    # the initial score
                 sound_params = {'direction':0.0},    # properties of the score response sound
                 gain_file = 'default_ding.wav',      # sound file per point
                 loss_file = 'default_buzz.wav',      # sound file for losses
                 none_file = 'default_click.wav',     # file to play if no reward
                 ding_interval = 0.1,                 # interval at which successive gain sounds are played... (if score is > 1)
                 buzz_volume = 0.1,                   # volume of the buzz (multiplied by the amount of loss)
                 gain_volume = 0.25,                  # volume of the gain sound
                 scorefile = 'scoretable.txt',        # this is where the scores are logged
                 engine = None                        # stimulus-presentation engine (the ._engine field of an instance of BasicStimuli)
                 ):
        BasicStimuli.__init__(self)
        if engine is not None:
            self._engine = engine
        self.score = initial_score
        self.params = sound_params
        self.gain_file = gain_file
        self.loss_file = loss_file
        self.none_file = none_file
        self.ding_interval = ding_interval
        self.buzz_volume = buzz_volume
        self.gain_volume = gain_volume
        self.scorefile = scorefile
        self.paused = False
    
        # open the score log
        with open(self.scorefile,'a') as score_log:
            score_log.write(time.asctime() + ': ----- new session -----\n')
    
    def score_event(self,delta,nosound=False):
        """Handle a score update."""
        if self.paused:
            return
        self.marker(150+delta)
        self.score = self.score+delta
        if not nosound:
            if delta>0:
                # play k gain sound events for a score delta of k
                self.marker(1)
                taskMgr.doMethodLater(0,self.play_gain,'Score sound')
                while delta > 1:
                    taskMgr.doMethodLater(delta*self.ding_interval,self.play_gain,'Score sound')
                    delta -= 1
            elif delta<0:
                # play k loss sound events for a score delta of -k
                self.marker(2)
                taskMgr.doMethodLater(0,self.play_loss,'Score sound')
                while delta < -1:
                    taskMgr.doMethodLater(-delta*self.ding_interval,self.play_loss,'Score sound')
                    delta += 1                    
            else:
                # play the no-score delta sounds (probably unused)
                self.sound(self.none_file,volume=self.buzz_volume,**self.params)

    def play_gain(self,task):
        self.sound(self.gain_file,volume=self.gain_volume,**self.params)
        self.marker(1)
        return task.done
 
    def play_loss(self,task):
        self.sound(self.loss_file,volume=self.buzz_volume,**self.params)
        self.marker(2)
        return task.done
    
    def log_score(self):    
        with open(self.scorefile,'a') as score_log:
            score_log.write(time.asctime() + ': ' + str(int(self.score*10)) + '\n')        
    