# -*- coding:utf-8 -*-
from framework.latentmodule import LatentModule
from direct.gui.OnscreenImage import OnscreenImage
from panda3d.core import Point2, TransparencyAttrib,Texture
import random
import math
import time 

brownian_rand = random.Random()


class BrownianItems(LatentModule):
    """Items that are moving according to Brownian motion."""
    def __init__(self,
                 
                 # item control
                 num_items=200,                                                 # number of items to maintain
                 item_graphics = ['tactical\\unit2.png','tactical\\unit5.png','tactical\\unit8.png',
                                  'tactical\\unit11.png','tactical\\unit12.png','tactical\\unit13.png',
                                  'tactical\\unit13.png','tactical\\unit19.png','tactical\\unit20.png',
                                  'tactical\\unit21.png'], # item graphics to use (randomly chosen from that set)
                 item_speed = lambda: brownian_rand.uniform(0,0.1),             # overall item movement speed; may be callable 
                 item_angle = lambda: brownian_rand.vonmisesvariate(0,0),       # overall item movement direction, in radians; may be callable
                 item_diffusion = lambda: brownian_rand.normalvariate(0,0.025),  # item Brownian perturbation process (applied at each frame); may be callable
                 item_size = lambda: brownian_rand.uniform(0.01,0.03),          # item size in arena coord system; may be callable
                 item_roll = 0,                                                 # item rotation in radians; may be callable
                 item_color = [1,1,1,1],                                        # the color of the item; may be callable
                 item_oriented = False,                                          # orient items in the direction of movement; may be callable
                 item_lifetime = 1000000,                                       # lifetime of the items (after which a new one would be respawned elsewhere); may be callable
                 item_spiral = None,                                            # perform a spiraling motion with the given radius and angular velocity

                 # arena control
                 arena = [-0.7,0.5,-0.7,0.5],               # the coordinate space in which the action takes place (left,right,bottom,top)
                 boundary = 0.2,                            # zone around the arena in which things can move around and spawn
                 repulsion = [0,2],                         # rebound force driving items aways from the screen boundaries; [constant term,linear term]
                 respawn_in = 'boundary',                   # where to spawn: can be 'boundary','arena','both'
                 zsort = 0,                                 # z sorting of the generated items
                 
                 # misc control
                 pixelated=False,                           # whether to show unfiltered (pixelated) images 
                 renderer=None,                             # the renderer which shall display everything
                 lifetime=None):                            # lifetime of this stream, in seconds
       
        LatentModule.__init__(self)
        self.num_items = num_items
        self.item_speed = item_speed
        self.item_angle = item_angle
        self.item_diffusion = item_diffusion
        self.item_size = item_size
        self.item_graphics = item_graphics
        self.item_roll = item_roll
        self.item_color = item_color
        self.item_oriented = item_oriented
        self.item_lifetime = item_lifetime        
        self.item_spiral = item_spiral
        self.arena = arena
        self.boundary = boundary
        self.repulsion = repulsion
        self.respawn_in = respawn_in
        self.zsort = zsort
        self.renderer = renderer
        self.lifetime = lifetime
        self.pixelated = pixelated
        
        self.items = []       # the array of item resources (incl display position)
        self.item_vel = []    # array of item current velocities
        self.item_drift = []  # array of item accelerations
        self.item_eol = []    # end of lifetime fo the icons
        self.item_rot = []    # rotation of the item relative to movement direction if oriented, otherwise None
        self.item_spi = []    # spiraling motion control
          
    def run(self):
        if self.lifetime is None:
            self.lifetime = 1000000
            
        try:
            # generate initial batch of items
            for k in xrange(self.num_items):
                self._spawn_item()
            # go to sleep keeping items updating
            self.sleep(self.lifetime,self._update)
        finally:
            # destroy items
            for k in xrange(self.num_items):
                self.items[k].removeNode()

        
    # spawn a new item (optionally at some index in the items array)
    def _spawn_item(self, idx=None):        
        # select the icon
        pic = brownian_rand.choice(self.item_graphics)
        
        # select the spawn position
        if idx is None or self.respawn_in == 'both':
            
            # either in the arena or boundary
            x = brownian_rand.uniform(self.arena[0]-self.boundary,self.arena[1]+self.boundary)
            y = brownian_rand.uniform(self.arena[2]-self.boundary,self.arena[3]+self.boundary)
            
        elif self.respawn_in == 'boundary':
            
            # sample until we're inside the boundary region
            while True:
                x = brownian_rand.uniform(self.arena[0]-self.boundary,self.arena[1]+self.boundary)
                y = brownian_rand.uniform(self.arena[2]-self.boundary,self.arena[3]+self.boundary)
                if not (self.arena[0]<x<self.arena[1] and self.arena[2]<y<self.arena[3]):
                    break
                
        else:
            
            # within the arena
            x = brownian_rand.uniform(self.arena[0],self.arena[1])
            y = brownian_rand.uniform(self.arena[2],self.arena[3])
            
        pos = [x,0,y]
        
        # select the size
        scale = self.item_size() if callable(self.item_size) else self.item_size
        # select the rotation
        rot = [0,0,self.item_roll() if callable(self.item_roll) else self.item_roll]
        # select the color
        col = self.item_color() if callable(self.item_color) else self.item_color
        # select the drift vector
        ang = self.item_angle() if callable(self.item_angle) else self.item_angle
        spd = self.item_speed() if callable(self.item_speed) else self.item_speed
        drift = Point2(spd*math.cos(ang),spd*math.sin(ang))
        eol =  time.time() + (self.item_lifetime() if callable(self.item_lifetime) else self.item_lifetime)
         
        # enlist the item
        item = OnscreenImage(image=pic,pos=pos,scale=scale,hpr=rot,color=col,parent=self.renderer,sort=self.zsort)        
        item.setTransparency(TransparencyAttrib.MAlpha)
        if self.pixelated:
            item.getTexture().setMagfilter(Texture.FTNearest)
            item.getTexture().setMinfilter(Texture.FTNearest)
        
        spiral = self.item_spiral() if callable(self.item_spiral) else self.item_spiral
        if not (self.item_oriented() if callable(self.item_oriented) else self.item_oriented):
            rot = None
        else:
            rot = rot[2]
        
        if idx is None:
            # append a new item
            self.items.append(item)
            self.item_vel.append(Point2(0,0))
            self.item_drift.append(drift)
            self.item_eol.append(eol)
            self.item_rot.append(rot)
            self.item_spi.append(spiral)
        else:
            # replace the item at idx
            self.items[idx].removeNode()
            self.items[idx] = item
            self.item_vel[idx] = Point2(0,0)
            self.item_drift[idx] = drift
            self.item_eol[idx] = eol
            self.item_rot[idx] = rot
            self.item_spi[idx] = spiral


    def _update(self,dt):
        now = time.time()        
        rad2deg = 180.0 / (math.pi/2)
        for k in xrange(self.num_items):            
            # generate new random impulse and add to velocity
            self.item_vel[k] = Point2(self.item_diffusion() if callable(self.item_diffusion) else self.item_diffusion,self.item_diffusion() if callable(self.item_diffusion) else self.item_diffusion) + self.item_vel[k]
            
            # update the position
            delta = (self.item_vel[k] + self.item_drift[k]) * dt
            pos = [self.items[k].getPos()[0] + delta[0],self.items[k].getPos()[2] + delta[1]]
            
            # optionally update the rotation!
            if self.item_rot[k] is not None:
                ang = math.atan2(self.item_vel[k][1], self.item_vel[k][0]) * rad2deg
                self.items[k].setHpr(0,0,ang + self.item_rot[k])

            if self.item_spi[k] is None:
                self.items[k].setPos(pos[0],self.zsort,pos[1])
            else:
                # update angle
                self.item_spi[k][0] += dt * self.item_spi[k][2]
                # update pos
                ofs = [math.cos(rad2deg*self.item_spi[k][0])*self.item_spi[k][1],
                       math.sin(rad2deg*self.item_spi[k][0])*self.item_spi[k][1]]
                self.items[k].setPos(pos[0]+ofs[0],self.zsort,pos[1]+ofs[1])
                                    
            # check if we need to respawn
            if not (self.arena[0]-self.boundary < pos[0] < self.arena[1]+self.boundary) or not (self.arena[2]-self.boundary < pos[1] < self.arena[3]+self.boundary):
                self._spawn_item(k)
            if now > self.item_eol[k]:
                self._spawn_item(k)
            