from framework.latentmodule import LatentModule
from framework.navigation import *
from panda3d.core import VBase4,Point3,NodePath,Camera
import math

global city_model
city_model = loader.loadModel('big\\citty.egg') 

global last_k
last_k = None

class CheckpointDriving(LatentModule):
    """
    Show a city from a driving car's perspective. The car moves through a sequence
    of checkpoints.
    """
    def __init__(self,
                 # display parameters 
                 presenter = None,              # the text presenter on which to display messages
                 frame = [0.1,0.4,0.25,0.75],   # the display region in which to draw everything
                 text_pos = [0,-0.3],           # position of the text display
                 focused = True,               # whether this task is currently focused
                 
                 # media                  
                 envmodel = 'big\\citty.egg',                # the environment model to use
                 navmesh = "media\\citty_navmesh.bin",  # the navigation mesh for the environment
                 trucksound = "diesel_loop.wav",        # loopable truck sound....
                 trucksound_volume = 0.03,              # volume of the sound
                 trucksound_direction = 0,              # direction relative to listener
                 target_model = "moneybag-rev.egg",     # model of the target object
                 target_scale = 0.01,                   # scale of the target model
                 target_offset = 0.2,                   # y offset for the target object 

                 # checkpoint logic
                 sourcefile='media\\checkpoints.txt',           # file containing checkpoints & target locations... 
                 snap_radius = 30,                              # radius for snapping to nearest environmental point (if outside environment), in meters
                 points = [[-248.91,-380.77,4.812],[0,0,0]],    # the sequence of destination points...
                 radius = 2,                                    # proximity to checkpoint at which it is considered reached... (meters)
                 pathoffset = 1,                                # offset into the path at which we start

                 # target object positions
                 targets = [[0,0,0]],

                 # end conditions
                 end_timeout = 100000,          # end the task after this time
                 # movement parameters
                 acceleration = 0.5,            # acceleration during manual driving
                 friction = 0.95,               # friction coefficient
                 torque = 1,                    # actually angular velocity during turning
                 height = 0.7,                  # height of the camera over the ground
                 show_checkpoints = False,      # whether to show when a checkpoint is reached
                 ):

        LatentModule.__init__(self)
        
        self.presenter = presenter
        self.frame = frame
        self.envmodel = envmodel
        self.navmesh = navmesh
        self.trucksound = trucksound
        self.trucksound_volume = trucksound_volume
        self.trucksound_direction = trucksound_direction        
        self.target_model = target_model
        self.target_scale = target_scale 
        self.target_offset = target_offset
        self.points = points
        self.radius = radius
        self.acceleration = acceleration
        self.friction = friction
        self.torque = torque
        self.height = height
        self.pathoffset = pathoffset
        self.snap_radius = snap_radius     
        self.end_timeout = end_timeout
        self.text_pos = text_pos
        self.focused = focused
        self.show_checkpoints = show_checkpoints

        if sourcefile is not None:
            # read positions from a file
            self.targets = []
            self.points = []
            for line in open(sourcefile):
                if line.find('#') == -1:
                    star = line.find('*') 
                    if star >= 0:
                        self.targets.append(eval('[' + line[star+1:].strip() + ',0.0]'))
                    else:
                        tmp = eval('[' + line.strip() + ',0.0]')
                        self.points.append(tmp)
                    

    def run(self):
        try:
            self.accept('drive-report',self.onreport)
            if self.show_checkpoints:
                self.text = self.write('',10000,block=False,scale=0.04,wordwrap=25,pos=[self.text_pos[0],self.text_pos[1]-0.05])
            
            global city_model
            # create the display
            self.make_viewport()            
            
            self.write('one moment...',0.1,scale=0.04,pos=self.text_pos)
            # load the environment
            print "Loading environment; this may take a while..."            
            self.env = city_model #loader.loadModel(self.envmodel)
            self.env.reparentTo(self.render3d)
            
            # start engine sound
            self.engine = self.sound(self.trucksound,looping=True,playrate=1,volume=self.trucksound_volume,direction=self.trucksound_direction)
            
            # load navmesh and make a navcrowd
            self.nav = NavMesh(navmesh=self.navmesh)            
            self.crowd = NavCrowd(self.nav)
            
            # insert an object at each target location...
            self.bags = []
            for t in self.targets:
                bag = loader.loadModel(self.target_model)
                bag.reparentTo(self.render3d)
                pnt = detour2panda(self.nav.nearest_point(t,self.snap_radius)[1])
                bag.setPos(pnt[0],pnt[1],pnt[2]+self.target_offset)
                bag.setScale(self.target_scale,self.target_scale,self.target_scale)
                self.bags.append(bag)

            # add an agent at the starting point
            global last_k
            k = self.pathoffset if last_k is None else last_k
            point = self.nav.nearest_point(self.points[k%len(self.points)],self.snap_radius)
            self.crowd.add_agent(point)
            
            # cycle through the checkpoints
            t_end = time.time() + self.end_timeout
            while True:
                k += 1
                last_k = k
                target = self.points[k%len(self.points)]

                # head towards it
                self.crowd.replan_move_target(0, self.nav.nearest_point(target,self.snap_radius))
                                
                # ... and keep walking until it has been reached (or timed out)
                self.target = Point3(target[0],target[1],target[2])
                self.sleep(t_end - time.time(), self.update)

                if self.show_checkpoints:
                    self.write('Reached checkpoint ' + str(k) + ' (' + str(target[0]) + ',' + str(target[1]) +')',2,block=False,scale=0.04,wordwrap=25,pos=self.text_pos)
                
                if time.time() > t_end:
                    break
        finally:
            # free resources
            if hasattr(self,'bags'):
                for b in self.bags:
                    b.removeNode()
            try:
                self.engine.stop()
                self.render3d.removeNode()
                base.win.removeDisplayRegion(self.region)            
                self.ignoreAll()
            except:
                pass
        
    def update(self,dt):        
        agent = self.crowd.agent_status(0)
        
        # update camera position and velocity
        pos = agent.npos; pos[2] += self.height
        vel = agent.vel
        self.cam3d.setPos(pos)
        if vel.length() > 0:
            self.cam3d.lookAt(Point3(pos + vel))

        if self.show_checkpoints:
            self.text.setText('(' + str(pos[0]) + ',' + str(pos[1]) +')')
            
        # update engine sound play rate...
        self.engine.setPlayRate(1+vel.length()*0.1)
        
        # check reach condition
        distance = math.sqrt((pos[0]-self.target[0])*(pos[0]-self.target[0]) + (pos[1]-self.target[1])*(pos[1]-self.target[1]))
        if distance < self.radius:  
            self.resume()

    def make_viewport(self):
        # init the 2d display region and associated camera and renderer
        self.region = base.win.makeDisplayRegion(self.frame[0],self.frame[1],self.frame[2],self.frame[3])
        self.region.setClearColor(VBase4(0.6, 0.5, 1, 1))
        self.region.setClearColorActive(True)
        self.region.setClearDepthActive(True)
        self.cam3d = NodePath(Camera('cam3d'))
        self.render3d = NodePath('render3d')
        self.cam3d.reparentTo(self.render3d)
        self.region.setCamera(self.cam3d)
        return self.render3d

    def onreport(self):
        self.marker(903)
        self.sound('click2.wav',volume=1,direction=0,surround=True)
