from __future__ import with_statement
from framework.latentmodule import LatentModule
from framework.ui_elements.EventWatcher import EventWatcher
import random
import time

class CommScheduler(LatentModule):
    """
    A stream of stimuli with pseudo-random ISI out of a predefined stimulus set.
    If a stimulus contains the special assigned callsign, a "target" response is expected from the user.
    The stream is targeted at some presenter which presents the generated stimuli.
    """    
    def __init__(self,
                 # output environment
                 presenter,                                 # stimulus presenter to use
                 rewardlogic,                               # the reward handler                 
                 focused=True,                             # whether this object is in the user's focus
                 markerbase=300,                            # base marker offset
                 
                 # content control           
                 targetsign=None,                           # the user's assigned callsign (if None, will be randomly selected)
                 callsigns='callsigns.txt',                 # source file containing a list of call signs
                 numcallsigns=6,                            # subset of callsigns to use                 
                 commands='random_commands.txt',            # source file containing a list of commands
                 distractors='distractor_sentences.txt',    # source file containing a list of distractor sentences

                 # probabilities & timing control
                 cue_probability = 0.1,                     # probability that a cue message is selected
                 distractor_probability = 0.4,              # probability that a distractor message is selected
                 nontarget_probability = 0.3,               # probability that a non-target callsign message is selected
                 target_probability = 0.2,                  # probability that a target callsign message is selected                 
                 isi = lambda: random.uniform(4,6),         # message interval
                                  
                 # end conditions
                 end_trials=100000,                         # number of trials to produce
                 end_timeout=100000,                        # lifetime of this stream, in seconds
                 
                 # response control
                 response_event='x',                        # response button to use
                 timeout = 3,                               # response timeout...                    
                 loss_nontarget=-1,                         # amount of loss incurred when pressing for a non-target or if not focused
                 loss_missed=-1,                            # amount of loss incurred when missing a target
                 gain_target=2,                             # amount of reward gained when spotting a target
                 ):
               
        LatentModule.__init__(self)
        self.presenter = presenter
        self.rewardlogic = rewardlogic
        self.focused = focused
        self.markerbase = markerbase
        self.targetsign = targetsign
        # load callsigns
        self.callsigns = []
        with open('media\\'+callsigns,'r') as f:
            for line in f:
                self.callsigns.append(line.strip())
        self.callsigns = self.callsigns[:numcallsigns]               
        # load commands
        with open('media\\'+commands,'r') as f:
            self.commands = f.readlines()
        # load distractors
        with open('media\\'+distractors,'r') as f:
            self.distractors = f.readlines()
        self.cue_probability = cue_probability
        self.distractor_probability = distractor_probability             
        self.nontarget_probability = nontarget_probability
        self.target_probability = target_probability
        self.isi = isi

        self.end_trials = end_trials
        self.end_timeout = end_timeout

        self.response_event = response_event
        self.timeout = timeout
        self.loss_nontarget=loss_nontarget
        self.loss_missed=loss_missed
        self.gain_target=gain_target
        self.control = False

    def run(self):        
        # pick a random callsign if necessary
        if self.targetsign is None:
            signidx = random.choice(xrange(len(self.callsigns)))
            self.marker(self.markerbase+20+signidx)
            self.targetsign = self.callsigns[signidx];
            
        # remove the targetsign from the callsigns
        self.callsigns.remove(self.targetsign)

        self.accept('control',self.oncontrol,[True])
        self.accept('control-up',self.oncontrol,[False])
        watcher = EventWatcher(eventtype=self.response_event,
                               handleduration=self.timeout,
                               defaulthandler=self.false_detection)
        
        self.sleep(self.isi())
        t_end = time.time() + self.end_timeout
        self.iscued = False            
        for trial in xrange(self.end_trials):
            if not self.focused:
                self.iscued = False
            duration = self.isi()
            # pick a message
            r = random.random()
            if r < self.distractor_probability:
                # display a distractor
                self.marker(self.markerbase+1)
                self.presenter.submit_wait(random.choice(self.distractors),self)
            elif r < self.distractor_probability+self.cue_probability:
                # display a cue
                self.marker(self.markerbase+2)
                self.presenter.submit_wait(self.targetsign + ', next time press control.',self)
                if self.focused:
                    self.iscued = True
            elif r < self.distractor_probability+self.cue_probability+self.nontarget_probability:
                # display a non-target
                self.marker(self.markerbase+3)
                self.presenter.submit_wait(self.substitute(random.choice(self.commands),random.choice(self.callsigns)),self)
            else:
                # display a target
                self.marker(self.markerbase+4 if self.focused else self.markerbase+5)
                # ... and expect a response
                if self.focused:
                    watcher.watch_for(self.correct, self.timeout, self.missed)
                self.presenter.submit_wait(self.substitute(random.choice(self.commands),self.targetsign),self)

            # wait for the ISI and scan for a response
            self.sleep(duration)
            # check termination conditions
            if time.time() > t_end:
                break

    def substitute(self,command,callsign):
        """Substitute a callsign into a command."""
        if command.find(' *')>0:
            # placeholder not at the beginning
            command.replace('*',callsign.lower())
        elif command.find('*')>0:
            # placeholder at the beginning
            command.replace('*',callsign)
        else:
            # no placeholder, prepend callsign
            command = callsign + ', ' + command[0].lower() + command[1:]
        return command
                 
    def oncontrol(self,status):
        self.control = status

    def missed(self):
        if self.focused:
            self.marker(self.markerbase+6)
            self.rewardlogic.score_event(self.loss_missed)
            self.iscued = False

    def false_detection(self):
        self.marker(self.markerbase+7)
        self.rewardlogic.score_event(self.loss_nontarget)
        
    def correct(self):        
        if self.iscued:
            self.marker(self.markerbase+8 if self.control else self.markerbase+9)
        else:
            self.marker(self.markerbase+10 if self.control else self.markerbase+11)
        if self.control == self.iscued:
            # the user correctly spots the warning event            
            self.rewardlogic.score_event(self.gain_target)
        else:
            # the user spotted it, but didn't get the cue right
            self.rewardlogic.score_event(self.loss_nontarget)
        self.iscued = False
        