from framework.latentmodule import LatentModule
from framework.ui_elements.TextPresenter import TextPresenter
from direct.gui.DirectGui import DirectButton
import random
import time

class MathScheduler(LatentModule):
    """
    A class that presents random math problems to a presenter and processes user input.
    Optionally, the math problems can be just a type of distractor stream. 
    """
    def __init__(self,
                 # facilities used by this object
                 rewardhandler=None,            # a RewardLogic instance that manages the processing of generated rewards/penalties 
                 presenter=None,                # the presenter on which to output the math problems
                 presenter_params={'pos':lambda:[random.uniform(-0.5,0.5),random.uniform(-0.5,0)],      # parameters of a textpresenter, if no presenter is given
                                   'clearafter':3,'framecolor':[0,0,0,0],'scale':0.1,'align':'center'},
                 focused = True,                # whether this task is currently focused

                 # end conditions
                 end_timeout=None,              # end presentation after the timeout has passed
                 end_numproblems=None,          # end presentation after this number of problems have been presented
                 
                 # stimulus presentation statistics
                 difficulty=1,                  # difficulty level of the problems
                 problem_interval = lambda: random.uniform(3,12), # delay before a new problem appears after the previous one has been solved
                 
                 # response timing
                 response_timeout=10.0,         # time within which the subject may respond to a problem
                 gain_correct=3,                # gain if problem solved correctly
                 loss_incorrect=-2,             # loss if problem solved incorrectly
                 
                 # parameters for the numpad
                 numpad_topleft=[1.4,-0.15],    # top-left corner of the numpad
                 numpad_gridspacing=[0.21,-0.21], # spacing of the button grid
                 numpad_buttonsize=[1,1],       # size of the buttons
                 numpad_textscale=0.2,       # size of the buttons
                 
                 passive=False                  # paassively display math problems without handling response
                 ):
                
        LatentModule.__init__(self)
        
        self.rewardhandler = rewardhandler
        self.presenter = presenter
        self.presenter_params = presenter_params
        self.end_timeout = end_timeout
        self.end_numproblems = end_numproblems
        
        self.difficulty = difficulty
        self.problem_interval = problem_interval
        self.response_timeout = response_timeout
        self.gain_correct = gain_correct
        self.loss_incorrect = loss_incorrect

        self.numpad_topleft = numpad_topleft
        self.numpad_gridspacing = numpad_gridspacing
        self.numpad_buttonsize = numpad_buttonsize
        self.numpad_textscale = numpad_textscale
        self.passive = passive
        self.focused = focused
        
    def run(self):
        try:
            if self.presenter is None:
                self.presenter = TextPresenter(**self.presenter_params)
            
            font = loader.loadFont('arial.ttf')
            xoff = 0.0
            yoff = 0.35
            size = 0.5
            # create the numpad
            self.buttons = []
            for k in range(10):
                if k==9:
                    x,y,n = 0,3,0
                else:
                    x,y,n = k%3,k/3,k+1
                self.buttons.append(DirectButton(frameSize=(-size+xoff,size+xoff,-size+yoff,size+yoff),
                                                 pos=(self.numpad_topleft[0] + x*self.numpad_gridspacing[0],0,self.numpad_topleft[1] + y*self.numpad_gridspacing[1]),
                                                 text_font=font, text=str(n), scale=self.numpad_textscale, command=messenger.send, extraArgs=['num-' + str(n)],
                                                 rolloverSound=None, clickSound=None))
            # and add the "next" button
            self.buttons.append(DirectButton(frameSize=(-(size+0.013)*2/0.13*0.2,(size+0.013)*2/0.13*0.2, -0.45,1.07 ),
                                             pos=(self.numpad_topleft[0] + 1.5*self.numpad_gridspacing[0],0,self.numpad_topleft[1] + 3*self.numpad_gridspacing[1]+self.numpad_textscale*0.15),
                                             text_font=font, text="NEXT", scale=0.65*self.numpad_textscale, command=messenger.send, extraArgs=['num-next'],
                                             rolloverSound=None, clickSound=None))

            # begin record-keeping...
            problems = 0
            starttime = time.time()
                
            # for each problem...
            while True:
                # wait for the inter-problem interval
                self.sleep(self.problem_interval())
            
                # generate a new problem            
                op = random.choice(['+','*']) if self.difficulty > 3 else '+'
                A = int(random.uniform(7,7+self.difficulty*10))
                B = int(random.uniform(7,7+self.difficulty*10))            
            
                # present it
                self.presenter.submit_wait(str(A) + " " + op + " " + str(B) + " = ",self)

                if not self.passive:
                    # record the pressed digits (and resume() once next is pressed)
                    self.acceptOnce('num-next',self.resume)
                    nums = self.watchfor_multiple(['num-0','num-1','num-2','num-3','num-4','num-5','num-6','num-7','num-8','num-9'], self.response_timeout,list_only=True)
                
                    # convert them into a number and check if the typed-in answer is correct
                    input = ''.join([n[4] for n in nums])
                    input = int(input) if input != '' else 0
            
                    if self.focused:
                        if input == (A+B if op=='+' else A*B):
                            self.rewardhandler.score_event(self.gain_correct)
                        else:
                            self.rewardhandler.score_event(self.loss_incorrect)
                            
                # check end conditions
                problems += 1
                if self.end_numproblems is not None and problems > self.end_numnumproblems:
                    break
                if self.end_timeout is not None and time.time() > starttime + self.end_timeout:
                    break
            
        finally:
            for b in self.buttons:
                b.destroy()

