from framework.latentmodule import LatentModule
import random
import time

class TargetScheduler(LatentModule):
    """ 
    A class that schedules "target" and "nontarget" events in random (but timed) order across multiple
    presenters, and handles the user's responses & reward feedback
    """
    def __init__(self,
                 # facilities used by this object
                 eventwatcher=None,             # an EventWatcher instance that is used to coordinate the subject's response to targets
                 rewardhandler=None,            # a RewardLogic instance that manages the processing of generated rewards/penalties 
                 presenters=None,               # an array of presenters across which to display target/non-target events
                 
                 # end conditions
                 end_timeout=None,              # end presentation after the timeout has passed
                 end_numtargets=None,           # end presentation after this number of targets has been presented
                 end_numstimuli=None,           # end presentation after this number of stimuli has been presented
                 
                 # stimulus presentation statistics
                 stimulus_interval = lambda: random.uniform(0.5,4), # inter-stimulus interval, given as a function that returns a random number
                 target_probability=0.1,        # the probability that a given stimulus is a target (may also be a random function returning True or False)
                 target_distribution=None,      # if a stimulus is a target, this is the probabilty that it shows up at a given presenter - one number per presenter (by default uniform probability)
                 nontarget_distribution=None,   # if a stimulus is a non-target, this is the probabilty that it shows up at a given presenter - one number per presenter (by default uniform probability)
                 
                 # response timing
                 responsetime=1.0,              # time within which the subject may respond to a target
                 
                 # reward/penalty values
                 hit_values=None,               # an array of reward values for correctly responding to a target - one per presenter (may be numbers or strings) 
                 miss_values=None):             # an array of penalty values (negative numbers) missing a target - one per presenter (may be numbers or strings)
                
        LatentModule.__init__(self)
        
        # if unspecified the target & non-target distributions are taken as uniform
        if target_distribution is None:
            target_distribution = [1./len(presenters)]*len(presenters)
        if nontarget_distribution is None:
            nontarget_distribution = [1./len(presenters)]*len(presenters)
        # ... their length should equal the # of presenters
        if len(target_distribution) != len(presenters):
            raise Exception("There must be one probability value per presenter in target_probabilities.")
        if len(nontarget_distribution) != len(presenters):
            raise Exception("There must be one probability value per presenter in nontarget_probabilities.")
        # if unspecified, the hit & miss values are 0 by default
        if hit_values is None:
            hit_values = 0
        if miss_values is None:
            miss_values = 0 
        # hit & miss values might be specified as numbers or strings instead of lists; then, they are turned here into lists
        if type(hit_values) in (int,float,str):
            hit_values = [hit_values]*len(presenters)
        if type(miss_values) in (int,float,str):
            miss_values = [miss_values]*len(presenters)            
        if len(hit_values) != len(presenters): 
            raise Exception("The length of hit_values does not match the # of presenters.")
        if len(miss_values) != len(presenters): 
            raise Exception("The length of miss_values does not match the # of presenters.")

        # also make sure that the probabilities are correctly normalized
        target_distribution = [p/sum(target_distribution) for p in target_distribution]
        nontarget_distribution = [p/sum(nontarget_distribution) for p in nontarget_distribution]

        self.eventwatcher = eventwatcher
        self.rewardhandler = rewardhandler
        self.presenters = presenters
        self.stimulus_interval = stimulus_interval
        self.target_probability = target_probability if callable(target_probability) else (lambda: random.random() < target_probability)                
        self.target_distribution = target_distribution
        self.nontarget_distribution = nontarget_distribution
        self.end_timeout = end_timeout
        self.end_numtargets = end_numtargets
        self.end_numstimuli = end_numstimuli
        self.responsetime = responsetime
        self.hit_values = hit_values
        self.miss_values = miss_values
        

    def run(self):
        print 'TargetScheduler initializing.'
        
        # begin record-keeping...
        targets = 0
        stimuli = 0
        starttime = time.time()
                
        # for each trial...
        while True:
            # decide if it is a target
            if self.target_probability():
                # presenting a target -- select the presenter on which to show it
                p = self._sample_discrete(self.target_distribution)
                print "presenting target at",p
                self.presenters[p].submit_wait("target",self)
                self.marker(4)
                # handle response
                print "Registering event watcher...", time.time()
                self.eventwatcher.watch_for(lambda: self._correct_response(p),self.responsetime,lambda: self._miss_response(p))
                targets += 1
            else:
                # presenting a non-target -- select the presenter on which to show it 
                p = self._sample_discrete(self.nontarget_distribution)
                print "presenting non-target at",p 
                self.presenters[p].submit_wait("nontarget",self)                
                self.marker(8)
            stimuli += 1
            
            # wait for the inter-stimulus interval
            self.sleep(self.stimulus_interval())
            
            # check end conditions
            if self.end_numtargets is not None and targets > self.end_numtargets:
                break
            if self.end_numstimuli is not None and stimuli > self.end_numstimuli:
                break 
            if self.end_timeout is not None and time.time() > starttime + self.end_timeout:
                break
    
        print 'TargetScheduler done.'
    
    def _correct_response(self,p):
        print 'correct response: ', self.hit_values[p]
        self.rewardhandler.score_event(self.hit_values[p])
    
    def _miss_response(self,p):
        print 'missed target response: ', self.miss_values[p]
        self.rewardhandler.score_event(self.miss_values[p])
    
    def _sample_discrete(self,dist):
        """Sample a number from a discrete distribution like target_distribution or nontarget_distribution."""
        acc = 0.0
        r = random.random()
        for p in xrange(len(self.presenters)):
            acc = acc + self.target_distribution[p]
            if r < acc:
                break
        return p
