# -*- coding:utf-8 -*-
from framework.deprecated.controllers import BrownianItems
from framework.latentmodule import LatentModule
from framework.ui_elements.EventWatcher import EventWatcher
from panda3d.core import NodePath, Camera, OrthographicLens
from direct.gui.DirectGui import *
import random
import time

class VisualSearchTask(LatentModule):
    """
    Presents moving clutter items on top a background image, and a series of
    trials (situations), each of which may or may not involve a special target item appearing.
    """
     
    def __init__(self,
                 rewardlogic = None,                            # the reward logic to handle hits & misses
                 renderviewport = None,                         # optionally the viewport into which the map should render (can be generated by make_viewport())
                 focused = True,                                # whether this task is currently focused
                 
                 # map
                 background = 'satellite_baseline.png',         # background image to use
                 frame = [0.6,0.85,0.25,0.75],                  # the display region in which to draw everything
                 frame_boundary = 0.2,                          # (invisible) zone around the display region in which things can move around and spawn

                 # parameters of the target/non-target item processes
                 clutter_params = {'pixelated':True},           # parameters for the clutter process
                 target_params = {'pixelated':True,
                                  'num_items':1,
                                  'item_graphics':['tactical\\unit15.png','tactical\\unit15.png','tactical\\unit17.png']}, # parameters for the target process
                 intro_text = None,                             # optional text to introduce the task
                 
                 # situational control
                 target_probability = 0.5,                      # probability of a new situation being a target situation (vs. non-target situation) 
                 target_duration = lambda: random.uniform(3,6), # duration of a target situation
                 nontarget_duration = lambda: random.uniform(5,15),# duration of a non-target situation
                 
                 # end conditions
                 end_trials = 1000000,                          # number of situations to produce (note: this is not the number of targets)
                 end_timeout = 1000000,                         # lifetime of this stream, in seconds (the stream ends if the trials are exhausted)
                 
                 # response control
                 response_event = 't',                          # the event that is generated when the user presses the response button
                 loss_misstarget = -1,                          # the loss incurred by missing a target                 
                 loss_nontarget = -1,                           # the loss incurred by a false detection
                 gain_target = 2,                               # the gain incurred by correctly spotting a target
                 
                 button_params=None,  
                 ):      
                
        LatentModule.__init__(self)
        self.rewardlogic = rewardlogic
        self.renderviewport = renderviewport

        self.background = background
        self.frame = frame

        self.clutter_params = clutter_params
        self.target_params = target_params
        self.intro_text = intro_text
        
        self.target_probability = target_probability
        self.target_duration = target_duration
        self.nontarget_duration = nontarget_duration
                
        self.end_trials = end_trials
        self.end_timeout = end_timeout
        
        self.response_event = response_event
        self.loss_misstarget = loss_misstarget
        self.loss_nontarget = loss_nontarget
        self.gain_target = gain_target
        self.focused = focused
        self.button_params = button_params

    def run(self):
        try:
            # make the viewport if it doesn't yet exist
            if self.renderviewport is None:
                self.make_viewport()
            # load & set background image
            # bk = DirectButton(,parent=self.renderviewport)
            if self.button_params is not None:
                self.btn = DirectButton(parent=self.renderviewport,**self.button_params)
                self.btn.setBin('gui-popup',50) 
            self.bk = DirectFrame(image=self.background,scale=0.5,parent=self.renderviewport)        
            # init a visual clutter process
            self.launch(BrownianItems(renderer=self.renderviewport,**self.clutter_params))
            if self.intro_text is not None:
                self.write(self.intro_text,10,parent=self.renderviewport,scale=0.04,pos=[0,0.45])
            # set up and event watcher
            watcher = EventWatcher(eventtype = self.response_event,defaulthandler=self.non_target)
        
            # display a sequence of situations
            t_end = time.time() + self.end_timeout
            for t in xrange(self.end_trials):
                
                # generate a new situation                
                if random.random() < self.target_probability:
                    duration = self.target_duration()
                    
                    # display a target situation
                    if self.focused:
                        watcher.watch_for(self.hit_target, duration, self.miss_target)
                    self.launch(BrownianItems(lifetime=duration,renderer=self.renderviewport,**self.target_params))
                    self.sleep(duration)
                    
                else:
                    # non-target situation; nothing happens
                    duration = self.nontarget_duration()
                    self.sleep(duration)
                    
                if time.time() > t_end:
                    break;

        finally:
            watcher.destroy()
            if hasattr(self,'bk'):
                self.bk.removeNode()
            if hasattr(self,'btn'):
                self.btn.removeNode()

    def hit_target(self):
        self.rewardlogic.score_event(self.gain_target)
    
    def miss_target(self):
        # called when the subject missed a target
        if self.focused:
            if self.loss_misstarget != 0:
                self.rewardlogic.score_event(self.loss_misstarget)
    
    def non_target(self):
        self.rewardlogic.score_event(self.loss_nontarget)
    
    def make_viewport(self):
        # init the 2d display region and associated camera and renderer
        # stores the result in self.renderviewport
        region = base.win.makeDisplayRegion(self.frame[0],self.frame[1],self.frame[2],self.frame[3])
        region.setSort(20) 
        mycam2d = NodePath(Camera('mycam2d'))
        lens = OrthographicLens()
        lens.setFilmSize(1,1)
        lens.setNearFar(-1000,1000)
        mycam2d.node().setLens(lens)
        maprenderer = NodePath('maprenderer')
        maprenderer.setDepthTest(False)
        maprenderer.setDepthWrite(False)
        mycam2d.reparentTo(maprenderer)
        region.setCamera(mycam2d)
        self.renderviewport = maprenderer
        return maprenderer
    