# -*- coding:utf-8 -*-
from framework.latentmodule import LatentModule
import framework.detour
import framework.detourcrowd
from panda3d.core import *
import math

class BasicNavTask(LatentModule):
    def __init__(self,
                 presenter,                   # the text presenter on which to display messages
                 envmodel='big\\citty.egg',        # the environment model to use
                 frame=[0.1,0.4,0.25,0.75],   # the display region in which to draw everything
                 autopilot = True,            # whether the agent is automatically driven
                 acceleration=0.5,            # acceleration during manual driving
                 friction=0.95,               # friction coefficient
                 torque=1,                    # actually angular velocity during turning
                 trucksound="diesel_loop.wav" # loopable truck sound....
                 ):
        LatentModule.__init__(self)

        self.presenter = presenter
        self.frame = frame
        print "loading environment model...",
        self.env = loader.loadModel(envmodel)
        print "success."
        self.autopilot = autopilot
        self.acceleration = acceleration
        self.friction = friction
        self.torque = torque
        self.trucksound = trucksound

    def run(self):
        try:
            # create the display
            self._make_display()
            
            # set up mildly annoying engine sound
            self.engine = loader.loadSfx(self.trucksound)
            self.engine.setLoop(True)
            self.engine.setPlayRate(0.5)
            self.engine.setVolume(0.1)
            self.engine.play()
            
            # make the environment displayable
            self.env.reparentTo(self.render3d)
            
            # set up the correct camera position & direction
            self.cam3d.setPos(-248.91,-380.77,4.812)
            self.cam3d.lookAt(0,0,0)
            self.velocity = 0
                        
            # set up keyboard hooks
            self.accept('x',self.on_toggleautopilot)
            self.accept('r',self.cam3d.setPos,[-248.91,-380.77,4.812])
            self.accept('a',self.on_turnleft)
            self.accept('d',self.on_turnright)
            self.accept('a-repeat',self.on_turnleft)
            self.accept('d-repeat',self.on_turnright)
            self.accept('w',self.on_movefwd)
            self.accept('s',self.on_moveback)
            self.accept('w-repeat',self.on_movefwd)
            self.accept('s-repeat',self.on_moveback)

            # load navmesh
            print "loading navigation mesh...",
            mesh = framework.detour.dtLoadMesh("media\\citty_navmesh.bin")
            self.mesh = mesh
            print "success."
            
            # create a query filter...
            print "generating query filter...",
            filter = framework.detour.dtQueryFilter()
            self.filter = filter
            print "success."
            
            # init a query object so that we can do a bunch of geometry queries...
            print "initializing navigation mesh query...",
            query = framework.detour.dtNavMeshQuery()
            query.init(mesh,16384).disown()
            print "success."
            
            print "executing polygon range query...",
            src_polyref = framework.detour.new_uintp(1)
            src_point = framework.detour.dtPoint3(0,0,0)
            pos = framework.detour.dtPoint3(self.cam3d.getPos()[0],self.cam3d.getPos()[2],-self.cam3d.getPos()[1])
            query.findNearestPoly(pos,framework.detour.dtPoint3(5,5,5),filter,src_polyref,src_point).disown()
            
            dst_polyref = framework.detour.new_uintp(1)
            dst_point = framework.detour.dtPoint3(0,0,0)
            query.findNearestPoly(framework.detour.dtPoint3(0,0,0),framework.detour.dtPoint3(30,30,30),filter,dst_polyref,dst_point).disown()
            self.query = query
            print "success."
            
            print "initializing crowd...",
            crowd = framework.detourcrowd.dtCrowd()
            crowd.init(10,0.6,mesh)
            print "success."
                        
            print "adding agent...",
            params = framework.detourcrowd.dtCrowdAgentParams()
            params.radius = 2
            params.height = 2
            params.maxAcceleration = 6
            params.maxSpeed = 9.5
            params.collisionQueryRange = params.radius * 12
            params.pathOptimizationRange = params.radius * 30
            params.updateFlags = framework.detourcrowd.DT_CROWD_ANTICIPATE_TURNS # | DT_CROWD_OPTIMIZE_VIS | DT_CROWD_OPTIMIZE_TOPO | DT_CROWD_OBSTACLE_AVOIDANCE | DT_CROWD_SEPARATION 
            params.obstacleAvoidanceType = 3
            params.separationWeight = 1
            crowd.addAgent(src_point,params)            
            print "success."
            
            print "requesting move target...",
            crowd.requestMoveTarget(0,framework.detour.uintp_getitem(dst_polyref,0),dst_point)
            print "success."
            
            self.crowd = crowd
            self.debuginfo = framework.detourcrowd.dtCrowdAgentDebugInfo()

            self.sleep(10000,self.update)
            
        except Exception, inst:
            print "Error in navigation task:"
            print inst
        finally:
            # unlink the environment
            self.env.removeNode()
            self.engine.stop()
            self.render3d.removeNode()
            base.win.removeDisplayRegion(self.region)
        
    def update(self,dt):
        if self.autopilot:
            self.crowd.update(dt,self.debuginfo)
            agent = self.crowd.getAgent(0)
            get = framework.detour.floatp_getitem
            p = agent.npos
            v = agent.vel
            p = Point3(get(p,0),-get(p,2),get(p,1) + 0.7)
            v = Point3(get(v,0),-get(v,2),get(v,1))
            self.cam3d.setPos(p)
            if v.length() > 0:
                self.cam3d.lookAt(Point3(p+v))
            self.velocity = v.length()
                
        else:            
            # update velocity (friction)
            self.velocity = self.velocity * math.pow(self.friction, dt)

            # determine distance to nearest plane for height correction
            m = self.cam3d.getMat();
            pos = m.getRow3(3)
            p = framework.detour.dtPoint3(0,0,0); tmp_polyref = framework.detour.new_uintp(1)
            self.query.findNearestPoly(framework.detour.dtPoint3(pos[0],pos[2],-pos[1]),framework.detour.dtPoint3(5,5,5),self.filter,tmp_polyref,p).disown()
            offset = Point3(0,0,framework.detour.floatp_getitem(p,1)-pos[2] + 0.7)
            
            # update position    
            self.cam3d.setMat(m * m.translateMat(offset) * m.translateMat(Point3(m.getRow3(1)*self.velocity*dt)))        
                      
        # update engine sound play rate...
        self.engine.setPlayRate(0.5+self.velocity*0.1)
         
    def on_toggleautopilot(self):
        self.autopilot = not self.autopilot
        
    def on_turnleft(self):
        self.cam3d.setH(self.cam3d.getH()+self.torque);
        
    def on_turnright(self):
        self.cam3d.setH(self.cam3d.getH()-self.torque);
        
    def on_movefwd(self):
        self.velocity = self.velocity + self.acceleration 

    def on_moveback(self):
        self.velocity = self.velocity*0.75 #- self.acceleration
        
    def _make_display(self):
        # init the 2d display region and associated camera and renderer
        self.region = base.win.makeDisplayRegion(self.frame[0],self.frame[1],self.frame[2],self.frame[3])
        self.region.setClearColor(VBase4(0.6, 0.5, 1, 1))
        self.region.setClearColorActive(True)
        self.region.setClearDepthActive(True)
                
        self.cam3d = NodePath(Camera('cam3d'))
        self.render3d = NodePath('render3d')
        self.cam3d.reparentTo(self.render3d)
        self.region.setCamera(self.cam3d)


