# -*- coding:utf-8 -*-
from __future__ import with_statement
from functools import partial
import random
import time
from framework.deprecated.subtasks import StimulusStream

class CallsignStream(StimulusStream):
    """
    A stream of stimuli with pseudo-random ISI out of a predefined stimulus set.
    If a stimulus contains the special assigned callsign, a "target" response is expected from the user.
    
    The stream is targeted at some presenter which presents the generated stimuli.
    """
    
    def __init__(self,
                 presenter,                                 # the target presenter which should receive the messages (e.g. a TextPresenter)
                 targetsign=None,                           # the user's assigned callsign (if None, will be randomly selected)
                 callsigns='callsigns.txt',                 # source file containing a list of call signs                 
                 commands='random_commands.txt',            # source file containing a list of commands
                 trials=None,                               # number of trials to produce
                 lifetime=None,                             # lifetime of this stream, in seconds
                 isi=partial(random.uniform,4,6), # inter-stimulus interval; specified as a function that produces a positive random number
                 retry=True,                                # whether to retry submitting the distractor or drop it when the presenter was locked                
                 locktime=1.5,                              # time for which the message presenter shall be locked after a stimulus was emitted
                 minisi=3,                                  # the minimum isi (in case the random number is negative or so)
                 shuffle_commands=True,                     # whether to shuffle the source commands
                 numcallsigns=6):                           # maximum number of callsigns to use                               
        """Construct a CallsignStream."""
        StimulusStream.__init__(self)

        # load callsigns
        self.callsigns = []
        with open('media\\'+callsigns,'r') as f:
            for line in f:
                self.callsigns.append(line.strip())
        self.callsigns = self.callsigns[1:6]
               
        # load commands
        with open('media\\'+commands,'r') as f:
            self.commands = f.readlines()        
        
        # insert randomly selected callsigns  
        for k in xrange(len(self.commands)):
            s = self.commands[k]
            # pick one
            cs = self.callsigns[random.choice(xrange(len(self.callsigns)))]
            # insert
            if s.find(' *')>0:
                # placeholder not at the beginning
                s.replace('*',cs.lower())
            elif s.find('*')>0:
                # placeholder at the beginning
                s.replace('*',cs)
            else:
                # no placeholder, prepend callsign
                s = cs + ', ' + s[0].lower() + s[1:];         
            self.commands[k] = s
        
        # optionally permute the commands
        if shuffle_commands:
            random.shuffle(self.commands)

        # store parameters
        self.presenter = presenter
        self.targetsign = targetsign
        self.trials = trials
        self.lifetime = lifetime
        self.isi = isi 
        self.retry = retry
        self.locktime = locktime
        self.minisi = minisi
    
    def run(self):        
        try:
            starttime = time.time()
            
            # pick a random callsign if necessary
            if self.targetsign is None:
                self.targetsign = self.callsigns[random.choice(xrange(len(self.callsigns)))].lower();
                while True:
                    self.submit('Your callsign is '+self.targetsign+' -- please confirm with Enter.',self.presenter,3)
                    # wait for the user to confirm that he understood it
                    if self.waitfor('enter',5) is not None:
                        self.submit('Acknowledged. Callsign task begins now.',self.presenter)
                        break
                    
            trial = 0
            while True:
                # display a command
                command = self.commands[trial % len(self.commands)]
                istarget = command.lower().find(self.targetsign.lower())>=0
                success = self.submit(command,self.presenter,self.locktime,block=False)
                
                if success or not self.retry:
                    # successfully emitted the command: wait & check for a response
                    if len(self.watchfor('target-response',max(self.isi(),self.minisi))) == 0:
                        # no response
                        if istarget:
                            self.submit("You MISSED THE TARGET!",self.presenter,self.locktime)
                    else:
                        # responded
                        if istarget:
                            self.submit("Roger!",self.presenter)
                        else:
                            self.submit("THAT WAS NO TARGET!!",self.presenter)
                    trial = trial+1
                else:
                    # could not submit the message: retry
                    self.sleep(max(self.isi(),self.minisi))
                                
                # check termination conditions
                if (self.lifetime is not None) and (starttime + self.lifetime > time.time()):
                    break
                if (self.trials is not None) and (trial > self.trials):
                    break
        finally:
            pass
    