# -*- coding:utf-8 -*-
from __future__ import with_statement
from functools import partial
import random
import time
from framework.deprecated.subtasks import StimulusStream

class DistractorStream(StimulusStream):
    """
    This is a generic & reusable sub-task for any kind of randomly timed 
    distractor event presentation; works across any MessagePresenter subclass.
    """
    
    def __init__(self,
                 presenter,                                 # the target presenter which should receive the messages (e.g. a TextPresenter)
                 sourcefile='distractor_sentences.txt',     # source file containing a list of messages
                 shuffle=True,                              # whether to shuffle the source items
                 trials=None,                               # number of trials to produce
                 lifetime=None,                             # lifetime of this stream, in seconds
                 isi=partial(random.uniform,1,5),           # inter-stimulus interval; specified as a function that produces a positive random number or as a number
                 stimtime=None,                             # lifetime of the stimulus (until it will be cleared); if set, should be <locktime and <minisi
                 retry=True,                                # whether to retry submitting the distractor or drop it when the presenter was locked
                 locktime=0.75,                             # time for which the message presenter shall be locked after a stimulus was emitted
                 minisi=0.5):                               # the minimum isi
          
        """Construct a DistractorStream."""
        StimulusStream.__init__(self)

        # load items 
        with open('media\\'+sourcefile,'r') as f:
            self.items = f.readlines()
            
        # optionally permute the items
        if shuffle:
            random.shuffle(self.items)

        # store parameters
        self.presenter = presenter
        self.trials = trials
        self.lifetime = lifetime
        self.isi = isi 
        self.retry = retry
        self.locktime = locktime
        self.minisi = minisi
        self.stimtime = stimtime
            
    def run(self):        
        try:
            starttime = time.time()
            trial = 0
            while True:
                # produce an event
                success = self.presenter.submit(self.items[trial % len(self.items)],self.locktime, self.stimtime() if callable(self.stimtime) else self.stimtime)
                if success or not self.retry:
                    trial = trial+1
                    
                # wait for the isi
                self.sleep(max(self.isi() if callable(self.isi) else self.isi,self.minisi))
                                
                # check termination conditions
                if (self.lifetime is not None) and (starttime + self.lifetime > time.time()):
                    break
                if (self.trials is not None) and (trial > self.trials):
                    break
        finally:
            pass
