# -*- coding:utf-8 -*-
from __future__ import with_statement
from functools import partial
import random
import time
from framework.deprecated.subtasks import StimulusStream

class TargetStream(StimulusStream):
    """
    This is a generic & reusable sub-task for any kind of randomly timed 
    target event presentation; works across any MessagePresenter subclass.
    """
    
    def __init__(self,
                 presenter,                                 # the target presenter which should receive the messages (e.g. a TextPresenter)
                 sourcefile='target_sentences.txt',         # source file containing a list of messages
                 correct_response='CORRECT!',               # response to present if the user was correct
                 miss_response='YOU MISSED A NOTIFICATION TARGET!', # response to present if the user missed a target
                 shuffle=True,                              # whether to shuffle the source items
                 trials=100,                               # number of trials to produce
                 lifetime=None,                             # lifetime of this stream, in seconds
                 isi=partial(random.uniform,4,20),          # inter-stimulus interval; specified as a function that produces a positive random number or as a number
                 stimtime=2,                                # lifetime of the stimulus (until it will be cleared); if set, should be <locktime and <minisi
                 responsetime=2,                            # time window within which the user may respond
                 locktime=2):                               # time for which the message presenter shall be locked after a stimulus was emitted
          
        """Construct a TargetStream."""
        StimulusStream.__init__(self)

        # load items 
        with open('media\\'+sourcefile,'r') as f:
            self.items = f.readlines()
            
        # optionally permute the items
        if shuffle:
            random.shuffle(self.items)

        # store parameters
        self.presenter = presenter
        self.trials = trials
        self.lifetime = lifetime
        self.isi = isi 
        self.locktime = locktime
        self.stimtime = stimtime
        self.responsetime = responsetime
        self.correct_response = correct_response
        self.miss_response = miss_response
            
    def run(self):        
        try:
            starttime = time.time()
            for trial in xrange(self.trials):
                
                # produce an event
                self.submit(self.items[trial % len(self.items)],self.presenter,self.locktime,clearafter=self.stimtime() if callable(self.stimtime) else self.stimtime)
                # wait for a response 
                if self.waitfor('target-response',self.responsetime) is not None:
                    # user responded
                    self.submit(self.correct_response,self.presenter,1)
                else:
                    # user failed to respond
                    self.submit(self.miss_response,self.presenter,1)

                # wait for the rest of the ISI (TODO: handle stray target presses here)
                self.sleep((self.isi() if callable(self.isi) else self.isi)-self.responsetime)
                
                # check termination conditions
                if (self.lifetime is not None) and (starttime + self.lifetime > time.time()):
                    break
                if (self.trials is not None) and (trial > self.trials):
                    break
                
        finally:
            pass
