# -*- coding:utf-8 -*-
from __future__ import with_statement
from framework.latentmodule import LatentModule
import random
import time

class TextComprehensionStream(LatentModule):
    """
    A stream of events of the following type:
     - a text paragraph (or single line) is being presented
     - a series of yes/no questions is being presented, where the user is expected to
       respond with either yes or no responses
        
    The stimulus presentation takes place on a specific message presenter.
    
    The format of the source text file is as follows:
    
    Paragraph text body blah blah blah.
    Some question about the text? | Yes
    Another question about the text? | No
    
    Next paragraph text body blah blah blah...
    Some question about the text? | No
    
    Next paragraph text body blah blah blah...
    Some question about the text? | No
    Another question about the text? | Yes
    
    """
    
    def __init__(self,
                 presenter,                                 # the target presenter which should receive the messages (e.g. a TextPresenter)
                 data='paragraphs.txt',                     # source file containing a list of paragraphs with associated questions and correct answers                 
                 presenttime=3,                             # the waiting time between presentation of the material and before the the presenter will be cleared
                 presentlocktime=3,                         # time for which the message presenter shall be locked after the paragraph was presented 
                 clearedtime=2,                             # time between the presenter getting cleared and the first question appearing 
                 questiontime=3,                            # the time that the user has to respond to each question
                 questionlocktime=2,                        # time for which the message presenter shall be locked after a question was presented 
                 trials=None,                               # number of trials to produce
                 lifetime=None,                             # lifetime of this stream, in seconds
                 isi=lambda: random.uniform(1,5),           # inter-stimulus interval; specified as a function that produces a positive random number
                 retry=True,                                # whether to retry submitting the distractor or drop it when the presenter was locked
                 shuffle=True,                              # whether to shuffle the material to be presented
                 feedback_time=5,                           # duration for which the feedback is displayed               
                 minisi=0.5):                               # the minimum isi (in case the random number is negative or so)
        
        """Construct a TextComprehensionStream."""
        LatentModule.__init__(self)

        # load database
        self.data = []
        with open('media\\'+data,'r') as f:
            curitem = []
            for line in f:
                if len(line.strip()) == 0:
                    # blank line; finalize/append the item
                    self.data.append(curitem)
                    curitem = []
                else:
                    if len(curitem) == 0:
                        # first line; add in the paragraph body
                        curitem.append(line)
                    else:
                        # yes/no question + answer line; add the question (before the vertical bar) and the answer type (after the bar)
                        spl = line.split('|')
                        curitem.append((spl[0].strip(),spl[1].strip().lower()))
        
        # optionally permute the material
        if shuffle:
            random.shuffle(self.data)

        # store parameters
        self.presenter = presenter
        self.presenttime = presenttime
        self.presentlocktime = presentlocktime
        self.clearedtime = clearedtime
        self.questiontime = questiontime
        self.questionlocktime = questionlocktime
        self.trials = trials
        self.lifetime = lifetime
        self.isi = isi 
        self.retry = retry
        self.feedback_time = feedback_time
        
        self.minisi = minisi
    
    def run(self):        
        try:
            starttime = time.time()
            trial = 0
            while True:
                # get the next item
                item = self.data[trial % len(self.data)]
                # display the paragraph
                self.presenter.submit_wait(item[0],self,self.presentlocktime)
                # successfully submitted the command; wait for some time...
                self.sleep(self.presenttime)
                # clear the presenter
                self.presenter.clear()
                # and wait for some more time
                self.sleep(self.clearedtime)
                # then go through each of the questions
                for question in item[1:]:
                    # get the question presented
                    self.presenter.submit_wait(question[0],self,self.questionlocktime)                    
                    # wait for a response
                    resp = self.waitfor_multiple(['y','n'],self.questiontime) 
                    if resp is None:
                        # The question was missed; display a complaint
                        self.presenter.submit_wait("You failed to respond to a question!",self)
                    else:
                        # the user responded; check whether the response was correct or not
                        type = question[1]                            
                        if resp[0] == type[0]:
                            self.presenter.submit_wait("Correct!",self)
                        else:
                            self.presenter.submit_wait("Wrong!",self)
                    self.sleep(self.feedback_time)
                trial = trial + 1
                
                # wait for the isi duration
                self.sleep(max(self.isi(),self.minisi))

                # check termination conditions
                if (self.lifetime is not None) and (starttime + self.lifetime > time.time()):
                    break
                if (self.trials is not None) and (trial > self.trials):
                    break
        finally:
            pass
